# In order for the generated lexer to subclass our lexer base
# class, we have to add the lines
#       from generated.FoxySheepParser import FoxySheepParser
#       from FoxySheep.LexerBase import LexerBase
# to the generated lexer.

import sys

from antlr4 import Lexer
from FoxySheep.generated.InputFormParser import InputFormParser


class LexerBase(Lexer):
    def __init__(self, input_, output=sys.stdout):
        super(LexerBase, self).__init__(input_, output=output)
        self.bracketLevel = 0
        self.lastToken = None

    # Binary plus follows a complete expression. Complete
    # expressions always end with one of the following
    # tokens. On the other hand, unary plus never follows
    # these tokens. Distinguishing unary plus from binary
    # plus disambiguates the grammar and allows us to use
    # implicit multiplication.

    closeExprTokens = [
        InputFormParser.DecimalNumber,
        InputFormParser.NumberInBase,
        InputFormParser.Name,
        InputFormParser.StringLiteral,
        InputFormParser.RPAREN,
        InputFormParser.RBRACE,
        InputFormParser.HASH,
        InputFormParser.PERCENTDIGITS,
        InputFormParser.PERCENTS,
        InputFormParser.TRIPPLEBLANK,
        InputFormParser.DOUBLEBLANK,
        InputFormParser.BLANKDOT,
        InputFormParser.BLANK,
        InputFormParser.HASHDIGITS,
        InputFormParser.HASHStringLiteral,
        InputFormParser.DOUBLEHASHDIGITS,
        InputFormParser.HASH,
        InputFormParser.DOUBLEHASH,
        InputFormParser.DIGITS,
        InputFormParser.RBRACKET,
        InputFormParser.RBARBRACKET,
        InputFormParser.BoxRightBoxParenthesis,
        InputFormParser.DOUBLEPLUS,
        InputFormParser.DOUBLEMINUS,
        InputFormParser.BANG,
        InputFormParser.DOUBLEBANG,
        InputFormParser.CONJUGATE,
        InputFormParser.TRANSPOSE,
        InputFormParser.CONJUGATETRANSPOSE,
        InputFormParser.HERMITIANCONJUGATE,
        InputFormParser.SINGLEQUOTE,
        InputFormParser.DOUBLESEMICOLON,
        InputFormParser.DOUBLEDOT,
        InputFormParser.TRIPPLEDOT,
        InputFormParser.AMP,
        InputFormParser.DOT,
        InputFormParser.SEMICOLON,
    ]

    # Curiously, the lexer does not allow us to inspect previous
    # tokens. Thus we need to keep track of the previous token
    # so that we can use it to disambiguate unary/binary plus.
    # I don't think getToken consumes a token, so I don't think we need to record lastToken here.
    # def getToken(self):
    #     lt = super(LexerBase, self)._token
    #     if lt.channel != self.HIDDEN:
    #         self.lastToken = lt
    #     return lt
    def nextToken(self):
        lt = super(LexerBase, self).nextToken()
        if lt.channel != self.HIDDEN:
            self.lastToken = lt
        return lt

    # The following checks to see if the previous token likely
    # ended an expr. If so, it returns true. We use this method
    # in an action on plus to disambiguate between unary plus
    # and binary plus.
    def precededByExpr(self):
        # Returns true if the previous token ended a complete expr.
        if self.lastToken is None:
            return False
        tokenType = self.lastToken.type

        return tokenType in self.closeExprTokens

    # To determine if a newline separates expressions, we keep
    # track of the bracketing level. Note that we treat all
    # bracket-like characters as the same.
    def incrementBracketLevel(self, i):
        self.bracketLevel += i

    # We need to differentiate between ";;" when it follows a complete
    # expression or not in order to solve a context sensitivity
    # problem in the parser. See the Span parser rules for details.
    def checkDoubleSemicolon(self):
        if not self.precededByExpr():
            self.type = InputFormParser.SPANSEMICOLONS
            # self.setType(InputFormParser.SPANSEMICOLONS)

    # We distinguish between unary plus and binary plus.
    def checkAdditiveOp(self, type):
        if self.precededByExpr():
            self.type = type
            # self.setType(type)

    # Note that FoxySheep treats newlines the same way Mathematica does:
    # Wolfram Language "treats the input that you give on successive
    # lines as belonging to the same expression whenever no complete
    # expression would be formed without doing this."
    #
    # The following checks to see if the current token (a newline)
    # separates two expressions using the following heuristic:
    # If the token follows a complete expression and all bracket-
    # like characters have been matched, then the token is an
    # expression separator, and we return true.
    def checkNewline(self):
        if not (self.precededByExpr() and self.bracketLevel == 0):
            self._channel = self.HIDDEN
            # self.setChannel(self.HIDDEN)



"""
Classes for the type system. The difference between a Wolfram Language
type and a type in the target language can be blurry. Some design choices
have to be made about what types there should be.

The type of an AST node is dynamic in general and is computed during type
inference.
"""

from typing import List


# Since TreeTypeTrait is the only subclass modifying type comparison
# behavior, I don't need to make `is_subtype()` a static method of `TypeBase` so
# that it can be overridden by different subtypes as I did with `match_tree()`
# and `_match_node` in the `ASTNode` class.
def is_subtype(type_lhs:'TypeBase', type_rhs:'TypeBase') -> bool:
    """
    Determines if type_lhs is a subtype of type_rhs.

    :param type_lhs:
    :param type_rhs:
    :return: True if type_lhs is a subtype of type_rhs; otherwise, False.
    """

    # TODO: Implement type matching.
    pass


class TypeBase:

    @property
    def type_name(self):
        name = type(self).__name__

        # The type name is the name of the class without the suffix
        if name.endswith('Type'):
            return name[:-4]
        return name


class MathematicalTypeTrait:
    """
    This type is a mix-in that specifies the type trait that this type is a
    type of a mathematical expression.
    """
    pass


class NumberType(TypeBase, MathematicalTypeTrait):
    """
    This type includes floats and ints.
    """
    pass


class RealType(NumberType):
    pass


class IntegerType(NumberType):
    pass


class RationalType(NumberType):
    # Will this type ever arise in practice considering n/m will always be
    # parsed as division?
    pass


class SymbolType(TypeBase, MathematicalTypeTrait):
    """
    An identifier without a value and any other explicit or inferred type has
    type SymbolType. Do not confuse this type with the type of a SymbolNode,
    which may have any type.

    SymbolType plays the role of the most generic type. However, we are using
    the implementation language's (Python's) subclassing mechanism to
    implement both type traits, like MathematicalType, and type specificity,
    e.g. IntegerType is a NumberType. So the most generic type wants to be a
    subclass of every type trait class while simultaneously being a
    superclass of every type. The solution is that SymbolType is the only
    type that can be replaced with a non-subclass.
    """

    def __init__(self):
        super().__init__()


class TreeType:
    """
    This type is a mix-in that specifies the type trait that this type can
    appear as a non leaf node in a composite type represented by a tree
    structure.
    """

    def __init__(self, children: List["TypeBase"] = None,
                 parent: 'TreeType' = None, **kwargs):
        self.parent = parent
        self.children = children


class ListLikeType(TypeBase):
    """
    This class includes lists and associations.
    """
    pass



# Generated from InputForm.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .InputFormParser import InputFormParser
else:
    from InputFormParser import InputFormParser

# This class defines a complete generic visitor for a parse tree produced by InputFormParser.

class InputFormVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by InputFormParser#prog.
    def visitProg(self, ctx:InputFormParser.ProgContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#PatternExp.
    def visitPatternExp(self, ctx:InputFormParser.PatternExpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Or.
    def visitOr(self, ctx:InputFormParser.OrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Conjugate.
    def visitConjugate(self, ctx:InputFormParser.ConjugateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Ceiling.
    def visitCeiling(self, ctx:InputFormParser.CeilingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Infix.
    def visitInfix(self, ctx:InputFormParser.InfixContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Therefore.
    def visitTherefore(self, ctx:InputFormParser.ThereforeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#TagUnset.
    def visitTagUnset(self, ctx:InputFormParser.TagUnsetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Accessor.
    def visitAccessor(self, ctx:InputFormParser.AccessorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#CircleMinus.
    def visitCircleMinus(self, ctx:InputFormParser.CircleMinusContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Divide.
    def visitDivide(self, ctx:InputFormParser.DivideContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Implies.
    def visitImplies(self, ctx:InputFormParser.ImpliesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#PlusOp.
    def visitPlusOp(self, ctx:InputFormParser.PlusOpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#RightComposition.
    def visitRightComposition(self, ctx:InputFormParser.RightCompositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#NonCommutativeMultiply.
    def visitNonCommutativeMultiply(self, ctx:InputFormParser.NonCommutativeMultiplyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#List.
    def visitList(self, ctx:InputFormParser.ListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Cup.
    def visitCup(self, ctx:InputFormParser.CupContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Same.
    def visitSame(self, ctx:InputFormParser.SameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Optional.
    def visitOptional(self, ctx:InputFormParser.OptionalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SuchThat.
    def visitSuchThat(self, ctx:InputFormParser.SuchThatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#DoubleBracketingBar.
    def visitDoubleBracketingBar(self, ctx:InputFormParser.DoubleBracketingBarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#PatternBlankDot.
    def visitPatternBlankDot(self, ctx:InputFormParser.PatternBlankDotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Dot.
    def visitDot(self, ctx:InputFormParser.DotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#VerticalBar.
    def visitVerticalBar(self, ctx:InputFormParser.VerticalBarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Square.
    def visitSquare(self, ctx:InputFormParser.SquareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Alternatives.
    def visitAlternatives(self, ctx:InputFormParser.AlternativesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Out.
    def visitOut(self, ctx:InputFormParser.OutContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#BoxConstructor.
    def visitBoxConstructor(self, ctx:InputFormParser.BoxConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Not.
    def visitNot(self, ctx:InputFormParser.NotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Postfix.
    def visitPostfix(self, ctx:InputFormParser.PostfixContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#PatternBlanks.
    def visitPatternBlanks(self, ctx:InputFormParser.PatternBlanksContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#UnaryPlusMinus.
    def visitUnaryPlusMinus(self, ctx:InputFormParser.UnaryPlusMinusContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Cap.
    def visitCap(self, ctx:InputFormParser.CapContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#CirclePlus.
    def visitCirclePlus(self, ctx:InputFormParser.CirclePlusContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Because.
    def visitBecause(self, ctx:InputFormParser.BecauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#StringLiteral.
    def visitStringLiteral(self, ctx:InputFormParser.StringLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#And.
    def visitAnd(self, ctx:InputFormParser.AndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Get.
    def visitGet(self, ctx:InputFormParser.GetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Equivalent.
    def visitEquivalent(self, ctx:InputFormParser.EquivalentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#CompoundExpression.
    def visitCompoundExpression(self, ctx:InputFormParser.CompoundExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Derivative.
    def visitDerivative(self, ctx:InputFormParser.DerivativeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Slot.
    def visitSlot(self, ctx:InputFormParser.SlotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#RightTee.
    def visitRightTee(self, ctx:InputFormParser.RightTeeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Xor.
    def visitXor(self, ctx:InputFormParser.XorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Rule.
    def visitRule(self, ctx:InputFormParser.RuleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#HeadExpression.
    def visitHeadExpression(self, ctx:InputFormParser.HeadExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#ReplaceAll.
    def visitReplaceAll(self, ctx:InputFormParser.ReplaceAllContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Intersection.
    def visitIntersection(self, ctx:InputFormParser.IntersectionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#PreIncrement.
    def visitPreIncrement(self, ctx:InputFormParser.PreIncrementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Integrate.
    def visitIntegrate(self, ctx:InputFormParser.IntegrateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Set.
    def visitSet(self, ctx:InputFormParser.SetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#OpEquals.
    def visitOpEquals(self, ctx:InputFormParser.OpEqualsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Message.
    def visitMessage(self, ctx:InputFormParser.MessageContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Cross.
    def visitCross(self, ctx:InputFormParser.CrossContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#PatternTest.
    def visitPatternTest(self, ctx:InputFormParser.PatternTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Prefix.
    def visitPrefix(self, ctx:InputFormParser.PrefixContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Backslash.
    def visitBackslash(self, ctx:InputFormParser.BackslashContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Repeated.
    def visitRepeated(self, ctx:InputFormParser.RepeatedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#MapApply.
    def visitMapApply(self, ctx:InputFormParser.MapApplyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Union.
    def visitUnion(self, ctx:InputFormParser.UnionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#VerticalSeparator.
    def visitVerticalSeparator(self, ctx:InputFormParser.VerticalSeparatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Factorial.
    def visitFactorial(self, ctx:InputFormParser.FactorialContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SpanA.
    def visitSpanA(self, ctx:InputFormParser.SpanAContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Function.
    def visitFunction(self, ctx:InputFormParser.FunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Number.
    def visitNumber(self, ctx:InputFormParser.NumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Star.
    def visitStar(self, ctx:InputFormParser.StarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Comparison.
    def visitComparison(self, ctx:InputFormParser.ComparisonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#TagSet.
    def visitTagSet(self, ctx:InputFormParser.TagSetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Increment.
    def visitIncrement(self, ctx:InputFormParser.IncrementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#VerticalTilde.
    def visitVerticalTilde(self, ctx:InputFormParser.VerticalTildeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Colon.
    def visitColon(self, ctx:InputFormParser.ColonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SmallCircle.
    def visitSmallCircle(self, ctx:InputFormParser.SmallCircleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Parentheses.
    def visitParentheses(self, ctx:InputFormParser.ParenthesesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SpanB.
    def visitSpanB(self, ctx:InputFormParser.SpanBContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Condition.
    def visitCondition(self, ctx:InputFormParser.ConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Floor.
    def visitFloor(self, ctx:InputFormParser.FloorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Composition.
    def visitComposition(self, ctx:InputFormParser.CompositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#CircleDot.
    def visitCircleDot(self, ctx:InputFormParser.CircleDotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SymbolLiteral.
    def visitSymbolLiteral(self, ctx:InputFormParser.SymbolLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#CircleTimes.
    def visitCircleTimes(self, ctx:InputFormParser.CircleTimesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Unset.
    def visitUnset(self, ctx:InputFormParser.UnsetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Del.
    def visitDel(self, ctx:InputFormParser.DelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Diamond.
    def visitDiamond(self, ctx:InputFormParser.DiamondContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Wedge.
    def visitWedge(self, ctx:InputFormParser.WedgeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Put.
    def visitPut(self, ctx:InputFormParser.PutContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#StringJoin.
    def visitStringJoin(self, ctx:InputFormParser.StringJoinContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Tee.
    def visitTee(self, ctx:InputFormParser.TeeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SetContainment.
    def visitSetContainment(self, ctx:InputFormParser.SetContainmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Vee.
    def visitVee(self, ctx:InputFormParser.VeeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#CenterDot.
    def visitCenterDot(self, ctx:InputFormParser.CenterDotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Times.
    def visitTimes(self, ctx:InputFormParser.TimesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#StringExpression.
    def visitStringExpression(self, ctx:InputFormParser.StringExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#BracketingBar.
    def visitBracketingBar(self, ctx:InputFormParser.BracketingBarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Coproduct.
    def visitCoproduct(self, ctx:InputFormParser.CoproductContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#AngleBracket.
    def visitAngleBracket(self, ctx:InputFormParser.AngleBracketContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#Power.
    def visitPower(self, ctx:InputFormParser.PowerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#ContextName.
    def visitContextName(self, ctx:InputFormParser.ContextNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SimpleContext.
    def visitSimpleContext(self, ctx:InputFormParser.SimpleContextContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#CompoundContext.
    def visitCompoundContext(self, ctx:InputFormParser.CompoundContextContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#NumberBaseN.
    def visitNumberBaseN(self, ctx:InputFormParser.NumberBaseNContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#NumberBaseTen.
    def visitNumberBaseTen(self, ctx:InputFormParser.NumberBaseTenContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#numberLiteralPrecision.
    def visitNumberLiteralPrecision(self, ctx:InputFormParser.NumberLiteralPrecisionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#numberLiteralExponent.
    def visitNumberLiteralExponent(self, ctx:InputFormParser.NumberLiteralExponentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#OutNumbered.
    def visitOutNumbered(self, ctx:InputFormParser.OutNumberedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#OutUnnumbered.
    def visitOutUnnumbered(self, ctx:InputFormParser.OutUnnumberedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SlotDigits.
    def visitSlotDigits(self, ctx:InputFormParser.SlotDigitsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SlotNamed.
    def visitSlotNamed(self, ctx:InputFormParser.SlotNamedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SlotSequenceDigits.
    def visitSlotSequenceDigits(self, ctx:InputFormParser.SlotSequenceDigitsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SlotSequence.
    def visitSlotSequence(self, ctx:InputFormParser.SlotSequenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#SlotUnnamed.
    def visitSlotUnnamed(self, ctx:InputFormParser.SlotUnnamedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#ExpressionListed.
    def visitExpressionListed(self, ctx:InputFormParser.ExpressionListedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#AccessExpressionA.
    def visitAccessExpressionA(self, ctx:InputFormParser.AccessExpressionAContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by InputFormParser#AccessExpressionB.
    def visitAccessExpressionB(self, ctx:InputFormParser.AccessExpressionBContext):
        return self.visitChildren(ctx)



del InputFormParser


from FoxySheep.tree.AlternativeNode import AlternativeNode
from FoxySheep.tree.BlankAbstractNode import BlankAbstractNode
from FoxySheep.tree.BlankNode import BlankNode
from FoxySheep.tree.BlankNullSequenceNode import BlankNullSequenceNode
from FoxySheep.tree.BlankSequenceNode import BlankSequenceNode
from FoxySheep.tree.ConditionNode import ConditionNode
from FoxySheep.tree.DefaultNode import DefaultNode
from FoxySheep.tree.OptionalNode import OptionalNode
from FoxySheep.tree.PatternNode import PatternNode
from FoxySheep.tree.PatternTestNode import PatternTestNode
from FoxySheep.tree.RepeatedNode import RepeatedNode
from FoxySheep.tree.RepeatedNullNode import RepeatedNullNode
from FoxySheep.tree.OptionsPatternNode import OptionsPatternNode



"""
Contains utilities to match patterns in order to determine if a
declaration overwrites or specifies another, associate usages with
declarations, determine if a type is a subtype of another type, etc.

# `match(lhs_node, rhs_node)` and `match_pattern([lhs], [rhs])`

The `match(lhs_node, rhs_node)` function takes two ASTNodes and is a
convenience function that immediately calls `match_pattern([lhs], [rhs])`,
a function taking lists of ASTNodes with an optional argument specifying
which of lhs or rhs must generalize the other.

This module implements a simple recursive descent pattern recognizer for the
matcher. The match function is in the spirit of the Wolfram Language built-in
`MatchQ`, except this function returns an ordering of the trees if they
match, whereas `MatchQ` returns a boolean. Also, `MatchQ` is *asymmetric* at
least with respect to options: `OptionsPattern` needs to be on the rhs in
`MatchQ`. Since this function returns an ordering, it is *antisymmetric*.

AST trees are partially ordered with respect to pattern matching in Wolfram
Language. As a consequence of this partial ordering, we have, for example,
that `f[2, _]` and `f[_, 5]` are incomparable. We the partial ordering
induced on individual node can be described as follows:

    ("-->" = "is generalized by")

                    +--> __ --+
    nonpattern--> _ +         |--> __. --> ___ --> ___.
                    +--> _. --+
    Level: 0      1      2          3       4       5

Note that _. and __ in level 2 are NOT comparable.

This matcher DOES match Head criteria, as in f[x_Integer]. The head must
be a SymbolNode.

This matcher DOES NOT evaluate conditional criteria (uses of /;),
since evaluation of the conditional for a given input can only be
determined at runtime. However, the presence of a conditional
*specializes* the pattern relative to the same pattern but without the
conditional. If both lhs and rhs have a conditional, they must
themselves match.

DefaultValues, along with all other  \*Values, are not included in
the matching at all (in keeping with `MatchQ`'s behavior). Only the
default value's parent OptionalNode is involved in the match.


# find_bound_variables(signature: FunctionDeclarationNode)

Constructs a list of Symbols that are bound within the scope of a function.

"""

from enum import IntEnum, auto
from typing import List, Tuple

from FoxySheep.tree.symbol_node import SymbolNode
from FoxySheep.scoping import Scope, Symbol
from FoxySheep.tree import treeNode
from FoxySheep.tree.pattern import PatternNode, BlankAbstractNode
from FoxySheep.utils.misc import camel_to_snake


def get_bound_symbols(signature: treeNode, scope: Scope = None)->List[Symbol]:
    """
    Creates a list of symbols bound by the given function signature (a
    pattern). Signature should have the form `Head[arguments_pattern]`. The
    `Head` of signature is ignored. Note that a signature is the LHS of a
    function declaration in FoxySheep.

    If a `Scope` is given, each symbol is added to `scope` after it is
    created. The `scope` should be that of the FunctionDeclarationNode. It is
    the caller's responsibility to create `scope` correctly.

    This function recursively searches the tree rooted at signature,
    so variables buried in other structures will be found, e.g. the `x` in
    `f[{x_}]`.

    :param signature: The function signature to search.
    :param scope: The scope of the current function to which each symbol
    should be added.
    :return: A list of `Symbols`
    """

    symbol_list = []
    # Create a *copy* of `signature`'s list of children. Since arguments is a
    # stack, we reverse the list so that the leftmost child is on top.
    arguments = list(reversed(signature.children))

    # Recursively add symbols to symbol_list
    _find_bound_symbols(arguments, symbol_list)

    # Add each symbol to the scope.
    scope.add_symbols(symbol_list)

    return symbol_list


def _find_bound_symbols(arguments: List[treeNode], symbol_list: list):
    if not arguments:
        # Done.
        return

    # Get the next node.
    node = arguments.pop()

    # If we found a variable
    if type(node) == PatternNode and node.identifier:
        head = None # The default.
        # Check if there is a required Head.
        if node.pattern and isinstance(node.pattern, BlankAbstractNode):
            head = node.pattern.required_head
        symbol = Symbol(node.identifier, required_head = head)

    # If we found something with children:
    elif node.children:
        # Recurse to children before moving on.
        child_arguments = list(reversed(node.children))
        _find_bound_symbols(child_arguments, symbol_list)

    # Now continue with the rest of arguments.
    _find_bound_symbols(arguments, symbol_list)


class ComparisonValue(IntEnum):
    """
    Possible relations in a partially ordered set.
        Specifies/Less: <
        Equal: =
        Generalizes/Greater: >
        Incomparable/Unequal: cannot be compared.
    The matcher returns a ComparisonValue.
    """
    Less = Specializes = auto()
    LessOrEqual = SpecializesOrEqual = auto()
    Equal = auto()
    GreaterOrEqual = GeneralizesOrEqual = auto()
    Greater = Generalizes = auto()
    Unequal = Incomparable = auto()


# The partial ordering excluding `Optional`s.
_order = {BlankNode: 1, BlankSequenceNode: 2, BlankNullSequenceNode: 3}


def match(lhs: List[treeNode], rhs: List[treeNode], gt: bool = None):
    """
    The driver of the pattern matcher. Does dynamic dispatch according to
    the subclass of each treeNode.

    :param gt: True if the lhs > rhs, False if rhs > lhs, and None if
    undetermined.
    :param lhs: A list of treeNode instances in reverse order from how
    they appear in the tree.
    :param rhs: A list of treeNode instances in reverse order from how
    they appear in the tree.
    :return: A ComparisonValue.
    """

    # The implementation of this function is complicated by the fact that we
    # are determining *which* of lhs/rhs dominates the other

    # TODO: Incorporate symbol attributes in matching.
    # The symbol attributes `Flat`, `Orderless`, and `OneIdentity` should
    # influence pattern matching, but this is not implemented.

    # TODO: Check for OptionsPattern first.

    # Base cases.
    if not lhs:
        if not rhs:
            return ComparisonValue.Equal
        else:
            success, gt = _swap(gt, False)
            if success:
                return _flip_result(match(rhs, lhs, gt))
            else:
                # We couldn't swap rhs and lhs, so either lhs!=rhs or lhs<=rhs
                # AND rhs<=lhs. Either way, match fails.
                return ComparisonValue.Incomparable

    # Peek at the top node.
    lhs_node = lhs[-1]

    # PatternNode is used to bind a symbol (variable name) to a pattern.
    # We ignore names of patterns, so we unwrap any PatternNode we find,
    # caring only about the patterns they contain. In the pathological
    # case that the PatternNode contains no pattern, we leave it alone.
    if isinstance(lhs_node, PatternNode):
        if PatternNode.pattern:
            lhs_node = lhs[-1] = PatternNode.pattern

    # Check if we should swap lhs and rhs, i.e., set the value of lhs_dominates.
    if rhs:
        rhs_node = rhs[-1]
        if type(rhs_node) == OptionalNode:
            if type(lhs_node) == OptionalNode:
                # Should we do a _match_node here first?
                # Do the optional rhs first.
                _match_optional(lhs, rhs, gt, optional_on_lhs = False)
            else:
                # Try to swap.
                success, gt = _swap(gt, False)
                if success:
                    return _flip_result(match(rhs, lhs, gt))
                else:
                    return ComparisonValue.Incomparable
        if type(lhs_node) in _order:
            if type(rhs_node) in _order:
                if _order[type(lhs_node)] < _order[type(rhs_node)]:
                    # Try to swap.
                    success, gt = _swap(gt, False)
                    if success:
                        return _flip_result(match(rhs, lhs, gt))
                    else:
                        return ComparisonValue.Incomparable
            # Enforce gt/prohibit any swap.
            _swap(gt, True)

    # This is the lazy programmers way of doing a switch on only a subset of
    # node classes. If the node's class is one for which we have an matching
    # _match function, call the appropriate function by ( string) name which
    # is extracted from the node's class's name. (The switch turns into a
    # hash.) This way, if we need to implement a new submatch function,
    # we just implement it and it will be used automatically.
    method = _dynamic_dispatch(lhs_node)
    if method:
        return method(lhs, rhs, gt)

    # At this point, lhs[-1] is not a pattern, and if rhs[-1] exists,
    # it is also not a pattern. If rhs is empty, match fails.
    if not rhs:
        return ComparisonValue.Incomparable

    # Check for equality of the trees rooted at lhs_node and rhs_node.
    result = _match_node(lhs, rhs, gt)

    return result


def match_pattern(lhs: treeNode, rhs: treeNode):
    """
    Determines if the argument pattern with root node_lhs matches the
    argument pattern with root node_rhs.


    :param lhs: The root node of the tree on the left to be matched against.
    :param rhs: The root node of the tree on the right to be matched against.
    :param match_types: Boolean. Whether or not to also match the inferred
    types. NOT IMPLEMENTED.
    :return: A ComparisonValue: Specifies ("<"), Equals ("="), Generalizes (
    ">"), or Incomparable.
    """

    # The work of doing the pattern matching is delegated to the match
    # function which takes lists of treeNodes, not treeNodes themselves.
    return match([lhs], [rhs])


def _synthesize(lhs: int, rhs: int) -> int:
    """
    Takes two ComparisonValues coming from comparing different parts of a
    tree  and synthesizes a compatible ComparisonValue. The compatible
    value is the most restrictive.

        >  and <  --> !=
        >  and >  --> >
        >  and >= --> >
        <  and >  --> !=
        <  and == --> <
        <  and <= --> <
        <= and >=--> ==
         etc.

    :param lhs:
    :param rhs:
    :return:
    """

    # Shortcut if they're the same.
    if lhs == rhs:
        return lhs

    # WLOG, assume lhs > rhs.
    if rhs > lhs:
        tmp = rhs
        rhs = lhs
        lhs = tmp

    if lhs == ComparisonValue.Incomparable:
        return lhs

    if lhs <= ComparisonValue.Equal:
        return rhs
    elif lhs == ComparisonValue.GreaterOrEqual and \
        rhs >= ComparisonValue.LessOrEqual:
        return ComparisonValue.Equal
    elif rhs == ComparisonValue.Equal:
        return lhs

    return ComparisonValue.Incomparable


def _swap(gt: bool, to: bool) -> Tuple[bool, bool]:
    """
    Attempts to set gt while enforcing the constraint that gt can be set to
    False only once and then only if gt is None, and gt cannot be set back to
    True once set to False. The swap fails only if one of the following:
        gt is False
        gt != None and to==False
    Here is every case:

        gt   | to    | success | result
        -----+-------+---------+-------
        None | True  | True    | True
        None | False | True    | False <-- Only case gt and to != success
        True | True  | True    | True
        True | False | False   | True
        False| True  | False   | False
        False| False | False   | False <-- Only case to==success != result

    Returns a tuple (success, lhs_dominates) where success is a boolean
    indicating whether the swap was successful, and the value of
    lhs_dominates after the attempt.

    :param gt: Boolean. What we are attempting to change.
    :param to: Boolean. What we are trying to change it to.
    :return: Tuple(success, new gt value).
    """
    if gt is None and not to:
        # gt has not been set and to==False
        return True, False

    gt = bool(gt)
    success = gt and to

    if not to and not success:
        return False, False

    # Covers all other cases.
    return (gt and to), (to == success)


def _flip_result(result: int):
    """
    Convenience function that exchanges < for > if lhs_dominates is False.

    Since this function is applied to return values of match(rhs, lhs) to
    flip-flop < and <= if gt is false. Since gt can only be flipped once,
    a call with argument order match(rhs, lhs) can only occur once, and so we
    don't ever have to flip-flop the other direction.

    :param result: A ComparisonValue.
    :return: Corrected ComparisonValue.
    """

    if result == ComparisonValue.Greater:
        return ComparisonValue.Less
    elif result == ComparisonValue.GreaterOrEqual:
        return ComparisonValue.LessOrEqual
    elif result == ComparisonValue.Less:
        # A sanity check. If we are constrained to lhs =< rhs and we find
        # that lhs > rhs, then lhs and rhs are incomparable.
        return ComparisonValue.Incomparable
    elif result == ComparisonValue.LessOrEqual:
        # Another sanity check. If we are constrained to lhs =< rhs and
        # we find that lhs > rhs, then lhs == rhs.
        return ComparisonValue.Equal

    return result


def _dynamic_dispatch(node: treeNode):
    """
    Returns the function _match_node_class_name if exists and None otherwise.

    :param node: An instance of some treeNode subclass.
    :return: The function _match_node_name, or None if it doesn't exist.
    """

    # Instance method names are in snake_case.
    method_name = '_match_' + camel_to_snake(
            # Remove the "Node" from the class name.
            node.__class__.__name__[:-4]
            )
    return globals().get(method_name, None)


def _match_node(lhs: List[treeNode], rhs: List[treeNode], gt: bool):
    """
    Determines whether the node at the top of lhs is the same treeNode
    subclass as the node at the top of rhs. In the case of two `SymbolNode`s,
    the `SymbolNodes`s match if and only if their names match. If the nodes
    match, match is called on their children. If successful, the match is
    continued with lhs and rhs.

    This function assumes lhs and rhs have been preprocessed by match,
    so do not call this function directly. Instead, call match_pattern.

    :param lhs: The first of two node lists to be compared.
    :param rhs: The second of two node lists to be compared.
    :param gt:  "Greater Than," holds state of whether lhs > rhs.

    :param match_types: Boolean. If True, one of the nodes' `type` property
    class type must be either the same class type or a subclass of the
    other node's `type` property's class type. NOT IMPLEMENTED.

    :return: A ComparisonValue: Specifies ("<"), Equals ("="), Generalizes (
    ">"), or Incomparable.
    """

    # TODO: Implement type matching.
    # if match_types:
    #     handle_error(NotImplementedError_("Matching types", treeNode=lhs_node))


    # Peak the current nodes.
    lhs_node = lhs[-1]
    rhs_node = rhs[-1]

    # The caller prohibits lhs_node < rhs_node, so we assume lhs_node >=
    # rhs_node.
    if type(lhs_node) != type(rhs_node):
        return ComparisonValue.Incomparable

    # SymbolNodes have an additional requirement that identifiers must match.
    if type(lhs_node) == SymbolNode and lhs_node.identifier != rhs_node.identifier:
        return ComparisonValue.Incomparable

    # Nodes themselves match. Descend into their children. Note that we make
    # *copies* of each list of children.
    children_lhs = list(reversed(lhs_node.children)) if lhs_node.children \
        else []
    children_rhs = list(reversed(rhs_node.children)) if rhs_node.children \
        else []
    result = match(children_lhs, rhs_node.children, gt)
    # Don't bother matching the rest if already !=.
    if result == ComparisonValue.Incomparable:
        return result

    # Match the rest of the stack.
    lhs_node = lhs.pop()
    rhs_node = rhs.pop()
    rest_result = match(lhs, rhs, gt)
    if rest_result == ComparisonValue.Incomparable:
        # Failure. Rewind the node stack.
        lhs.append(lhs_node)
        rhs.append(rhs_node)
        return rest_result

    return _synthesize(ComparisonValue.Equal, rest_result)


def _match_optional(lhs: List[treeNode],
                    rhs: List[treeNode],
                    gt: bool,
                    optional_on_lhs: bool = True):
    """
    Tries to match the optional node on the top of the lhs stack. It does
    this by first trying a match with the node, and if it fails, trying again
    without the node.

    :param lhs:
    :param rhs:
    :param gt:
    :return:
    """
    # DefaultValues, along with all other  \*Values, are not included in
    # the matching at all (in keeping with `MatchQ`'s behavior). Only the
    # default value's parent OptionalNode is involved in the match.

    if optional_on_lhs:
        option_side = lhs
    else:
        option_side = rhs

    # Get the pattern out of the option_node.
    option_node = option_side.pop()
    pattern_node = option_node.pattern

    # Try first with the node.
    option_side.append(pattern_node)
    result = match(lhs, rhs, gt)
    if result != ComparisonValue.Incomparable:
        # It worked.
        return result

    # Now try without the node. Take it back off the lhs stack.
    option_side.pop()
    result = match(lhs, rhs, gt)
    if result == ComparisonValue.Incomparable:
        # Failure. Rewind the stack.
        option_side.append(option_node)

    return result


def _match_blank(lhs: List[treeNode], rhs: List[treeNode], gt: bool):
    if not rhs:
        return ComparisonValue.Incomparable

    lhs_node = lhs.pop()
    rhs_node = rhs.pop()

    result = match(lhs, rhs, gt)
    if result == ComparisonValue.Equal and type(rhs_node) != BlankNode:
        return ComparisonValue.Generalizes
    elif result == ComparisonValue.Incomparable:
        # Failure. Rewind the node stacks.
        lhs.append(lhs_node)
        rhs.append(rhs_node)

    return result


def _match_blank_sequence(lhs: List[treeNode], rhs: List[treeNode], gt: bool):
    # BlankSequence needs to match at least one node in rhs.
    if not rhs:
        return ComparisonValue.Incomparable

    # We automatically consume the rhs node.
    rhs_node = rhs.pop()
    # The rest is exactly like BlankNullSequence, so we replace the top of
    # lhs with a BlankNullSequence.
    lhs_node = lhs.pop()
    lhs.append(BlankNullSequenceNode())

    result = _match_blank_null_sequence(lhs, rhs, gt)
    if result in (ComparisonValue.Less, ComparisonValue.Incomparable):
        # Failure. Rewind the stack.
        lhs[-1] = lhs_node
        rhs.append(rhs_node)
        return ComparisonValue.Incomparable
    elif result == ComparisonValue.LessOrEqual:
        return ComparisonValue.Equal

    return result


def _match_blank_null_sequence(lhs: List[treeNode], rhs: List[treeNode], gt: bool):
    if not rhs:
        return ComparisonValue.Generalizes

    # `_match_blank_null_sequence` (and `_match_blank_sequence`) eats everything
    # it can while still successfully matching. This code is very similar to
    # `_match_Optional`, except `Optional` wraps a pattern object.

    # Try first to consume a node, leaving BlankNullSequence on the stack.
    rhs_node = rhs.pop()
    result = match(lhs, rhs, gt)
    if result != ComparisonValue.Incomparable:
        # It worked.
        return result

    # Now try without consuming a node and removing BlankNullSequence from
    # the stack.
    rhs.append(rhs_node)
    lhs_node = lhs.pop()
    result = match(lhs, rhs, gt)
    if result == ComparisonValue.Incomparable:
        # Failure. Rewind the stack.
        lhs.append(lhs_node)

    return result



"""
Utility functions to interact with Mathematica.
"""

import time
import pexpect
from FoxySheep.Errors import handle_error, ExternalExceptionError

# I'm using mathline here. Maybe there should be an option for WolframKernel.
cmd_mathline = 'mathline --mainloop false --inoutstrings false --usegetline ' \
      'true --prompt "$$ "'
expect_first = '$$ In[1]:= '
wait_to_close = 0.1  # 100 milliseconds, probably too long.
FullForm_wrapper = 'FullForm[HoldAllComplete[{code}]]'
FullForm_strip = 'HoldAllComplete['
child = None


def close_pexpect():
    global child, wait_to_close

    if child and child.isalive():
        child.sendline('Exit[]')

    # Wait for the process to end.
    time.sleep(wait_to_close)

    if child and child.isalive():
        child.close(force=True)


def open_pexpect():
    global child

    if child and child.isalive():
        return True

    try:
        child = pexpect.spawn(cmd_mathline, encoding = 'utf-8')
    except pexpect.exceptions.ExceptionPexpect as e:
        handle_error(ExternalExceptionError(e))

    child.expect_exact(expect_first)
    child.setecho(False)
    return True


def evaluate(code: str):
    """
    Evaluates a string in a Mathematica kernel.

    Example:
        >>> evaluate('Plus@@{1, 2, 3}')
        '6'

    :param code: A string containing Mathematica code.
    :return: The result of the evaluation as a string.
    """

    # Lazy instantiation of the pexpect process.
    child = _pexpect()
    child.sendline(code)
    child.expect_exact('$$ ')

    # Convert newlines.
    result = child.before.replace('\r\n', '\n')
    # Remove terminating newline.
    result = result.strip()

    return result


def _pexpect():
    global child

    if child is None or not child.isalive():
        open_pexpect()

    if child is None or not child.isalive():
        return None

    return child


def FullForm(code: str):
    """
    Evaluates `FullForm[HoldAllComplete[code]]` in a Mathematica kernel.

    Example:
        >>> FullForm('f[x_3, ___]')
        'f[Times[3, Pattern[x, Blank[]]], BlankNullSequence[]]'


    :param code: A WolframLanguage expression as a string.
    :return: The FullForm representation of code.
    """

    global FullForm_wrapper

    result = evaluate(FullForm_wrapper.format(code=code))

    # Unwrap the HoldAll[ ].
    if result[0:16] == 'HoldAllComplete[':
        return result[16:-1]

    # Something went wrong. Return everything.
    return result


import atexit
atexit.register(close_pexpect)

if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""

"""

from FoxySheep.tree.nodes import treeNode, SymbolNode

class FunctionCallNode(treeNode):
    """
    The base class of all tree node classes representing a function call.

    The symbol_node
    """

    def __init__(self, symbol_node: SymbolNode, **kwargs):
        super().__init__(**kwargs)

        self.FunctionType = None


class FunctionDeclarationNode(treeNode):
    """
    The base class of all tree node classes representing something that
    creates a DownValue. This node will represent the DownValue in the symbol
    table. The symbol_node is the Head of the declaration, that is, the name
    of the function being created. The symbol_node will point to a Symbol in
    the symbol table. The symbol will have a DownValue that points back to
    this FunctionDeclarationNode.

    Properties:
        `symbol_node`: The name (head) of the function, a SymbolNode.
        `inner_scope`: The scope created by the function, if any.
    """

    def __init__(self, symbol_node: SymbolNode = None, **kwargs):
        """
        Create a new function declaration.

        :param symbol_node: The name (head) of the function.
        :param kwargs: The arguments to treeNodeBase.
        """

        super().__init__(**kwargs)

        # The Head of the declaration. The symbol_node will point to a Symbol
        # in the symbol table. The symbol will have a DownValue that points
        # back to this FunctionDeclarationNode.
        self.symbol_node = symbol_node

        # The function need not have a scope if it takes no variable arguments.
        # The inner_scope will need to be created based on the construction
        # of this node.
        self.inner_scope = None

    @property
    def lhs(self):
        if self.children:
            return self.children[0]
        return None

    @lhs.setter
    def lhs(self, symbol_node: SymbolNode, arguments_pattern: treeNode = None):
        self.children[0] = symbol_node

    @property
    def rhs(self):
        try:
            return self.children[1]
        except:
            return None

    @rhs.setter
    def rhs(self, rhs_node: treeNode):
        if not self.children:
            self.children.append(None)
        self.children[1] = treeNode



def split_on_case(s: str) -> str:
    """
    Given a string in CamelCase, yields each word.

    This monstrosity is due to StackExchange user
    [jdavidls](https://stackoverflow.com/users/1850574), from

    :param s: String to split.
    :yield: Next work in the string.
    """

    # state bits:
    # 0: no yields
    # 1: lower yields
    # 2: lower yields - 1
    # 4: upper yields
    # 8: digit yields
    # 16: other yields
    # 32 : upper sequence mark

    si, ci, state = 0, 0, 0  # start_index, current_index
    for c in s:

        if c.islower():
            if state & 1:
                yield s[si:ci]
                si = ci
            elif state & 2:
                yield s[si:ci - 1]
                si = ci - 1
            state = 4 | 8 | 16
            ci += 1

        elif c.isupper():
            if state & 4:
                yield s[si:ci]
                si = ci
            if state & 32:
                state = 2 | 8 | 16 | 32
            else:
                state = 8 | 16 | 32

            ci += 1

        elif c.isdigit():
            if state & 8:
                yield s[si:ci]
                si = ci
            state = 1 | 4 | 16
            ci += 1

        else:
            if state & 16:
                yield s[si:ci]
            state = 0
            ci += 1  # eat ci
            si = ci
        # Print debugging:
        # print(' : ', c, bin(state))
    if state:
        yield s[si:ci]


def camel_to_snake(s, lower: bool = True) -> str:
    """
    Converts CamelCase to snake_case. By default, it converts to lowercase.
    Pass lower=False to preserve the original case.

    This function is idempotent.

    Example:
        >>> camel_to_snake('ThisIsAMethodToSplitStrings')
        'this_is_a_method_to_split_strings'
        >>> camel_to_snake('ThisIsAMethodToSplitStrings', lower=False)
        'This_Is_A_Method_To_Split_Strings'

    :param s:
    :param lower:
    :return:
    """
    if lower:
        return '_'.join(split_on_case(s)).lower()
    return '_'.join(split_on_case(s))



from distutils.core import setup
import os.path as osp
from setuptools import find_packages


def get_srcdir():
    filename = osp.normcase(osp.dirname(osp.abspath(__file__)))
    return osp.realpath(filename)


srcdir = get_srcdir()


def read(*rnames):
    return open(osp.join(srcdir, *rnames)).read()


long_description = read("README.rst") + "\n"
exec(read("FoxySheep/version.py"))

setup(
    name="FoxySheep",
    author="Robert Jacobson",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    description="Mathematica parser and translator",
    version=VERSION,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(),
    entry_points={"console_scripts": ["foxy-sheep = FoxySheep.__main__:main"]},
    install_requires=["antlr4-python3-runtime>=4.7,<4.8", "click >= 7.1.2", "astor", "PyYAML"],
    url="http://github.com/rocky/FoxySheep2",
)



"""
Utilities to retrieve data about language built-ins, e.g. built-in function
declarations or names of operator symbols (`+` -> `"Plus"`) and vice versa.

The current implementation fetches this data from Mathematica itself,
assuming Mathematica is available on the command line as "math".
"""

from enum import IntEnum, auto

import FoxySheep.Utils.Mathematica as mma


class WolframLanguageRecord(IntEnum):
    NameExists          = auto()
    ArgumentsPattern    = auto()
    LocalVariables      = auto()
    Options             = auto()
    Attributes          = auto()


_SyntaxInformation = """Module[{r, p, prop},
    prop = {prop};
    r = FilterRules[_SyntaxInformation[{name}], prop];
    If[r == {{}}, r,
        p = r[[1]][[2]];
        If[p === None, {{}},
            If[prop == "LocalVariables", Flatten[p], p]
        ]
    ]
]"""

_get_code = {
    WolframLanguageRecord.NameExists:
        'Names["{name}"]',
    WolframLanguageRecord.ArgumentsPattern:
        _SyntaxInformation.format(prop="ArgumentsPattern"),
    WolframLanguageRecord.LocalVariables:
        _SyntaxInformation.format(prop="LocalVariables"),
    WolframLanguageRecord.Options:
        'First /@ Options[{name}]',
    WolframLanguageRecord.Attributes:
        'Attributes[{name}]'
    }


def find_symbol(name: str):
    """
    Finds symbol for the string name.
    """

    # Ask Mathematica if it knows the name.
    if mma.evaluate(
            _get_code[WolframLanguageRecord.NameExists].format(name=name)
            ) == '{}':
        # Nope.
        return None

    # Get ArgumentsPattern
    arguments_pattern = mma.evaluate(
            _get_code[WolframLanguageRecord.NameExists].format(name=name)
            )

    # Get scope type, etc.
    pass


def get_operator_name(symbolic_rep: str) -> str:
    pass



"""A collection of base classes and utility objects for building an abstract
syntax tree representing a Mathematica program. """

from typing import List

from FoxySheep.foxy_types import TypeBase

class Symbol(object):
    pass

class SymbolNode(object):
    pass

class treeNode(object):
    """
    The base class of all tree node classes.
    """

    def __init__(
        self,
        children: List["treeNode"] = None,
        parent: "treeNode" = None,
        type_: TypeBase = None,
        exprs: list = None,
    ):
        """
        Creates a node of the tree.

        :param children: A list of treeNode objects that are children of this node.
        :param parent: The parent_scope treeNode of this node.
        :param exprs: A list of the ANTLR SyntaxTree objects that represent this node.
        :param type_: The type of the value held by this node.
        """

        self.children = children
        self.parent = parent

        # A list of the ANTLR SyntaxTree objects that represent this node.
        self.exprs = exprs

        # The type returned or represented by the node.
        self._type = type_

        # Subclasses can register names for their child nodes which can be
        # accessed with __getattr__.
        self._named_children: dict = None

    def __getattr__(self, key: str):
        # Implements a mechanism to have child nodes with names as
        # attributes. See _add_named_child().

        # Check to see if key is one of the named children.
        if not self._named_children or key not in self._named_children:
            raise AttributeError(key)

        # If key is the name of a named child, find its index within
        # self.children.
        index = self._named_children[key]

        # Check that this node has children.
        if not self.children:
            return None

        # Try to retrieve the child.
        try:
            return self.children[index]
        except IndexError:
            pass

        return None

    def __setattr__(self, key, value):
        # Implements a mechanism to have child nodes with names as
        # attributes. See _add_named_child().

        # Check to see if key is one of the named children.
        if not self._named_children or key not in self._named_children:
            # Nope, not one of ours.
            super().__setattr__(key, value)

        # Since key is the name of a named child, find its index within
        # self.children.
        index = self._named_children[key]

        # Check that this node has children.
        if not self.children:
            self.children = []
        # Check that key's index exists.
        if len(self.children) < index + 1:
            # This index in self.children doesn't exist yet. We just extend
            # self.children with a list of None's.
            amount = index + 1 - len(self.children)
            self.children.extend(amount * [None])

        # Set the value.
        self.children[index] = value

    def _add_named_child(self, name: str, index: int):
        """
        Subclasses can register names for children at given indices in
        self.children so that a child with a name can be accessed by
        treeNode.name.

        :param name: String, the name of the child.
        :param index: Integer, the index in self.children at which the child
        lives.
        :return:
        """
        # Lazy instantiation.
        if not self._named_children:
            self._named_children = dict()
        self._named_children[name] = index

    def _add_named_children(self, names: List[str]):
        """
        Adds named children, automatically indexing starting at 0.

        :param names: List[str], the names of the children in order.
        :return:
        """
        for n, name in enumerate(names):
            self._add_named_child(name, n)

    def specify_type(self, type_: TypeBase):
        """
        The word *specify* is intentional, as type_ must be a subtype of this
        node's type unless this node's type is SymbolType.

        :param type_:
        :return:
        """
        # TODO: Implement treeNode.specify_type()
        pass

    @property
    def type_(self):
        return self._type



from FoxySheep.tree.nodes import treeNode, Symbol


class SymbolNode(treeNode):
    """
    A SymbolNode is a particular appearance in the source code of a
    name.

    A SymbolNode is not an entry in a symbol table. The Symbol that this
    SymbolNode points to is managed by the symbol table.

    In the case that the SymbolNode represents a usage of the symbol with no
    prior declaration, the SymbolNode gets its own (empty) symbol,
    and subsequent uses of the symbol will refer to this SymbolNode as their
    definition until the creation of a new definition.

    Properties:
        `identifier` String, the name of the symbol.
        `declaration` Where the symbol was declared.
    """

    def __init__(self, identifier: str, symbol: Symbol = None, **kwargs):
        """
        Create a SymbolNode representing an appearance of identifier in the
        source language. If a Symbol is provided, this SymbolNode's Symbol
        is the same type instance as declaration.

        :param identifier:
        :param UpValue:
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.identifier = identifier
        self._symbol = symbol

    @property
    def symbol(self):
        return self._symbol

    @symbol.setter
    def symbol(self, symbol: Symbol):
        assert symbol.identifier == self.identifier
        self._symbol = symbol



from AST import ASTNode, SymbolNode


class PatternNode(ASTNode):
    """
    `Pattern[identifier_, pattern_]`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._add_named_children(['identifier', 'pattern'])



"""
Driver routines for lexers and parsers

ANTLR4 is used to generate lexers and parsers. Here we glue lexing and parsing together.

"""
from antlr4 import InputStream, CommonTokenStream
from FoxySheep.generated.InputFormLexer import InputFormLexer
from FoxySheep.generated.InputFormParser import InputFormParser
from FoxySheep.generated.FullFormLexer import FullFormLexer
from FoxySheep.generated.FullFormParser import FullFormParser
from FoxySheep.transform.if_transform import input_form_post
from FoxySheep.emitter.full_form import input_form_to_full_form
from FoxySheep.emitter.python import input_form_to_python

# Cache the utility objects.
parser = None
lexer = None
ff_parser = None
ff_lexer = None
emitter = None

# FIXME: post_process is needed to flatten out binary opeators when going from
# InputForm to FullForm. It must not be used when going to Python.
# Probably the post_processing should be done outside of this function,
# and put into something more OutputForm specific.
def parse_tree_from_string(input_form: str, show_tree_fn=None):

    # Boilerplate
    # lexer = InputFormLexer(InputStream(input))
    # stream = CommonTokenStream(lexer)
    # parser = InputFormParser(stream)

    global parser, lexer

    # Reuse any existing parser or lexer.
    if not lexer:
        lexer = InputFormLexer(InputStream(input_form))
    else:
        lexer.inputStream = InputStream(input_form)
    if not parser:
        parser = InputFormParser(CommonTokenStream(lexer))
    else:
        parser.setTokenStream(CommonTokenStream(lexer))

    # Parse!
    tree = parser.prog()
    if show_tree_fn:
        show_tree_fn(tree, parser.ruleNames)
    return tree


def parse_tree_from_string_pp(input_form: str, show_tree_fn=None):
    tree = parse_tree_from_string(input_form, show_tree_fn)
    post_tree = input_form_post(tree)
    if post_tree != tree:
        show_tree_fn(post_tree, parser.ruleNames)
        tree = post_tree
    return tree


def ff_parse_tree_from_string(input: str, post_process=True, show_tree_fn=False):
    global ff_parser, ff_lexer

    # Reuse any existing parser or lexer.
    if not ff_lexer:
        ff_lexer = FullFormLexer(InputStream(input))
    else:
        ff_lexer.InputStream = InputStream(input)
        ff_lexer.reset()
    if not ff_parser:
        stream = CommonTokenStream(ff_lexer)
        ff_parser = FullFormParser(stream)
    else:
        # Don't create a new parser. Just reset the input stream.
        ff_parser.setTokenStream(CommonTokenStream(ff_lexer))
        ff_parser.reset()

    # Parse!
    tree = ff_parser.prog()
    if post_process:
        if show_tree_fn:
            show_tree_fn(tree, ff_parser.ruleNames)
        post_tree = input_form_post(tree)
        if post_tree != tree:
            show_tree_fn(post_tree, parser.ruleNames)
            tree = post_tree
    return tree

def if2ff(s: str, show_tree_fn=None) -> str:
    return input_form_to_full_form(s, parse_tree_from_string_pp, show_tree_fn)


def if2python(s: str, show_tree_fn=None) -> str:
    return input_form_to_python(s, parse_tree_from_string, show_tree_fn)


if __name__ == "__main__":
    print(if2ff("x^2-3x+4"))
    print(if2python("x^2-3x+4"))
